#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from ..characteristics import Characteristic
from ..bluetoothdevice import BluetoothDevice, ByteData
from typing import Union, Literal
from dataclasses import dataclass


AccelerometerPeriod = Union[
    Literal[1], Literal[2], Literal[5], Literal[10], Literal[20], Literal[80], Literal[160], Literal[640]
]


@dataclass
class AccelerometerData:
    x: int
    y: int
    z: int

    @staticmethod
    def from_bytes(values: ByteData):
        return AccelerometerData(
            int.from_bytes(values[0:2], "little", signed=True),
            int.from_bytes(values[2:4], "little", signed=True),
            int.from_bytes(values[4:6], "little", signed=True)
        )


class AccelerometerService:
    def __init__(self, device: BluetoothDevice):
        self._device = device

    def notify(self, callback):
        self._device.notify(Characteristic.ACCELEROMETER_DATA,
                            lambda sender, data: callback(AccelerometerData.from_bytes(data)))

    def read(self) -> AccelerometerData:
        return AccelerometerData.from_bytes(self._device.read(Characteristic.ACCELEROMETER_DATA))

    def set_period(self, period: AccelerometerPeriod):
        self._device.write(Characteristic.ACCELEROMETER_PERIOD, period.to_bytes(2, "little"))

    def read_period(self) -> int:
        return int.from_bytes(self._device.read(Characteristic.ACCELEROMETER_PERIOD)[0:2], "little")




#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from ..characteristics import Characteristic
from ..bluetoothdevice import BluetoothDevice


class LedService:
    def __init__(self, device: BluetoothDevice):
        self._device = device



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from ..characteristics import Characteristic
from ..bluetoothdevice import BluetoothDevice, ByteData

PDU_BYTE_LIMIT = 20


class UartService:
    def __init__(self, device: BluetoothDevice):
        self._device = device

    def receive(self, callback):
        self._device.notify(Characteristic.TX_CHARACTERISTIC, lambda sender, data: callback(data))

    def receive_string(self, callback):
        self.receive(UartService.to_string(callback))

    def send(self, data: ByteData):
        for i in range(0, len(data), PDU_BYTE_LIMIT):
            self._device.write(Characteristic.RX_CHARACTERISTIC, data[i:i + PDU_BYTE_LIMIT])

    def send_string(self, string: str):
        self.send(UartService.from_string(string))

    @staticmethod
    def from_string(string: str) -> bytes:
        return string.encode("utf-8")

    @staticmethod
    def to_string(callback):
        return lambda data: callback(str(data, "utf-8"))



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.
from typing import Callable
from enum import IntEnum
from ..characteristics import Characteristic
from ..bluetoothdevice import BluetoothDevice

ButtonCallback = Callable[[str], None]


class ButtonState(IntEnum):
    NOT_PRESSED = 0
    PRESS = 1
    PRESS_LONG = 2


class ButtonService:
    def __init__(self, device: BluetoothDevice):
        self._device = device

    @staticmethod
    def _create_button_callback(button: str, press, long_press, up):
        def button_callback(sender, data):
            if ButtonState.NOT_PRESSED == data[0] and up:
                up(button)
            elif ButtonState.PRESS == data[0] and press:
                press(button)
            elif ButtonState.PRESS_LONG == data[0] and long_press:
                long_press(button)
            else:
                pass
        return button_callback

    def on_button_a(self, press: ButtonCallback = None, long_press: ButtonCallback = None, up: ButtonCallback = None):
        self._device.notify(Characteristic.BUTTON_A,
                            ButtonService._create_button_callback('A', press, long_press, up))

    def on_button_b(self, press: ButtonCallback = None, long_press: ButtonCallback = None, up: ButtonCallback = None):
        self._device.notify(Characteristic.BUTTON_B,
                            ButtonService._create_button_callback('B', press, long_press, up))

    def read_button_a(self):
        return ButtonState(self._device.read(Characteristic.BUTTON_A)[0])

    def read_button_b(self):
        return ButtonState(self._device.read(Characteristic.BUTTON_B)[0])



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from enum import Enum


class Service(Enum):
    """
    Lists all services that could be offered by the microbit

    source: https://lancaster-university.github.io/microbit-docs/resources/bluetooth/bluetooth_profile.html
    """

    GENERIC_ACCESS = '00001800-0000-1000-8000-00805F9B34FB'
    """
    The generic_access service contains generic information about the device.
    All available Characteristics are readonly.
    """

    GENERIC_ATTRIBUTE = '00001801-0000-1000-8000-00805F9B34FB'
    """
    """

    DEVICE_INFORMATION = '0000180a-0000-1000-8000-00805f9b34fb'
    """
    The Device Information Service exposes manufacturer and/or vendor information about a device.            
    
    This service exposes manufacturer information about a device.
    The Device Information Service is instantiated as a Primary Service.
    Only one instance of the Device Information Service is exposed on a device.
    """

    ACCELEROMETER = 'E95D0753-251D-470A-A062-FA1922DFA9A8'
    """
    Exposes accelerometer data. An accelerometer is an electromechanical device that will measure acceleration forces. 
    These forces may be static, like the constant force of gravity pulling at your feet, or they could be dynamic 
    - caused by moving or vibrating the accelerometer.

    Value contains fields which represent 3 separate accelerometer measurements for X, Y and Z axes 
    as 3 unsigned 16 bit values in that order and in little endian format. 

    Data can be read on demand or notified periodically.
    """

    MAGNETOMETER = 'E95DF2D8-251D-470A-A062-FA1922DFA9A8'
    """
    Exposes magnetometer data.  A magnetometer measures a magnetic field such as the earth's magnetic field in 3 axes.
    """

    BUTTON = 'E95D9882-251D-470A-A062-FA1922DFA9A8'
    """
    Exposes the two Micro Bit buttons and allows 'commands' associated with button state changes to be associated 
    with button states and notified to a connected client.
    """

    IO_PIN = 'E95D127B-251D-470A-A062-FA1922DFA9A8'
    """
    Provides read/write access to I/O pins, individually or collectively. Allows configuration of each pin for 
    input/output and analogue/digital use.
    """

    LED = 'E95DD91D-251D-470A-A062-FA1922DFA9A8'
    """
    Provides access to and control of LED state. 
    Allows the state (ON or OFF) of all 25 LEDs to be set in a single write operation. 
    Allows short text strings to be sent by a client for display on the LED matrix and scrolled across at a speed 
    controlled by the Scrolling Delay characteristic.
    """

    EVENT = 'E95D93AF-251D-470A-A062-FA1922DFA9A8'
    """
    A generic, bi-directional event communication service. 

    The Event Service allows events or commands to be notified to the micro:bit by a connected client and it allows 
    micro:bit to notify the connected client of events or commands originating from with the micro:bit. The micro:bit 
    can inform the client of the types of event it is interested in being informed about (e.g. an incoming call) and 
    the client can inform the micro:bit of types of event it wants to be notified about.  
    
    The term “event” will be used here for both event and command types of data.
    
    Events may have an associated value.
    
    Note that specific event ID values including any special values such as those which may represent wild cards are 
    not defined here. The micro:bit run time documentation should be consulted for this information.
    
    Multiple events of different types may be notified to the client or micro:bit at the same time.
    Event data is encoded as an array of structs each encoding an event of a given type together with an associated 
    value. Event Type and Event Value are both defined as uint16 and therefore the length of this array will always be 
    a multiple of 4.
    
    struct event {
     uint16 event_type;
     uint16 event_value;
    };
    """

    DFU_CONTROL = 'E95D93B0-251D-470A-A062-FA1922DFA9A8'
    """
    Allows clients to initiate the micro:bit pairing and over the air firmware update procedures.
    """

    TEMPERATURE = 'E95D6100-251D-470A-A062-FA1922DFA9A8'
    """
    Ambient temperature derived from several internal temperature sensors on the micro:bit
    """

    UART = '6E400001-B5A3-F393-E0A9-E50E24DCCA9E'
    """
    This is an implementation of Nordic Semiconductor's UART/Serial Port Emulation over Bluetooth low energy. 

    See https://developer.nordicsemi.com/nRF5_SDK/nRF51_SDK_v8.x.x/doc/8.0.0/s110/html/a00072.html for the original 
    Nordic Semiconductor documentation by way of background.
    """



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from ..characteristics import Characteristic
from ..bluetoothdevice import BluetoothDevice


class TemperatureService:
    def __init__(self, device: BluetoothDevice):
        self._device = device

    def notify(self, callback):
        self._device.notify(Characteristic.TEMPERATURE, lambda sender, data: callback(
            int.from_bytes(data[0:1], 'little', signed=True)))

    def read(self) -> int:
        return int.from_bytes(self._device.read(Characteristic.TEMPERATURE)[0:1], 'little', signed=True)

    def set_period(self, period: int):
        self._device.write(Characteristic.TEMPERATURE_PERIOD, period.to_bytes(2, "little"))

    def read_period(self) -> int:
        return int.from_bytes(self._device.read(Characteristic.TEMPERATURE_PERIOD)[0:2], "little")



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from enum import Enum


class Characteristic(Enum):
    """
    Lists all characteristics in the microbit GATT profile

    source: https://lancaster-university.github.io/microbit-docs/resources/bluetooth/bluetooth_profile.html
    """

    DEVICE_NAME = '00002A00-0000-1000-8000-00805F9B34FB'
    """
    Read Mandatory
    Write Mandatory
    Fields
    1. Name : utf8s
    """

    APPEARANCE = '00002A01-0000-1000-8000-00805F9B34FB'
    """
    The external appearance of this device. The values are composed of a category (10-bits) and sub-categories (6-bits).
    
    Read Mandatory
    Fields
    1. Category : 16bit
    """

    PERIPHERAL_PREFERRED_CONNECTION_PARAMETERS = '00002A04-0000-1000-8000-00805F9B34FB'
    """
    Read Mandatory
    Fields
    1. Minimum Connection Interval : uint16
    2. Maximum Connection Interval : uint16
    3. Slave Latency : uint16
    4. Connection Supervision Timeout Multiplier : uint16
    """

    SERVICE_CHANGED = '2A05'
    """
    Indicate Mandatory
    Fields
    1. Start of Affected Attribute Handle Range : uint16
    2. End of Affected Attribute Handle Range : uint16
    Descriptors
    1. Client Characteristic Configuration : 2902
    """

    MODEL_NUMBER_STRING = '00002A24-0000-1000-8000-00805F9B34FB'
    """
    The value of this characteristic is a UTF-8 string representing the model number assigned by the device vendor.
    
    Read Mandatory
    Fields
    1. Model Number : utf8s
    """

    SERIAL_NUMBER_STRING = '00002A25-0000-1000-8000-00805F9B34FB'
    """
    The value of this characteristic is a variable-length UTF-8 string representing the serial number for a particular
    instance of the device.
    
    Read Mandatory
    Fields
    1. Serial Number : utf8s
    """

    HARDWARE_REVISION_STRING = '00002A27-0000-1000-8000-00805F9B34FB'
    """
    The value of this characteristic is a UTF-8 string representing the hardware revision for the hardware within 
    the device.
    
    Read Mandatory
    Fields
    1. Hardware Revision : utf8s
    """

    FIRMWARE_REVISION_STRING = '00002A26-0000-1000-8000-00805F9B34FB'
    """
    The value of this characteristic is a UTF-8 string representing the firmware revision for the firmware within 
    the device.
    
    Read Mandatory
    Fields
    1. Firmware Revision : utf8s
    """

    MANUFACTURER_NAME_STRING = '00002A29-0000-1000-8000-00805F9B34FB'
    """
    The value of this characteristic is a UTF-8 string representing the name of the manufacturer of the device.  
    
    Read Mandatory
    Fields
    1. Manufacturer Name : utf8s    
    """

    ACCELEROMETER_DATA = 'E95DCA4B-251D-470A-A062-FA1922DFA9A8'
    """
    Contains accelerometer measurements for X, Y and Z axes as 3 signed 16 bit values in that order and in little 
    endian format. X, Y and Z values should be divided by 1000.
    
    Read Mandatory
    Notify Mandatory
    Fields
    1. Accelerometer_X  : sint16
    2. Accelerometer_Y : sint16
    3. Accelerometer_Z : sint16
    Descriptors
    1. Client Characteristic Configuration : 2902
    """

    ACCELEROMETER_PERIOD = 'E95DFB24-251D-470A-A062-FA1922DFA9A8'
    """
    Determines the frequency with which accelerometer data is reported in milliseconds.
    Valid values are 1, 2, 5, 10, 20, 80, 160 and 640.
    
    Read Mandatory
    Write Mandatory
    Fields
    1. Accelerometer_Period : uint16
    """

    MAGNETOMETER_DATA = 'E95DFB11-251D-470A-A062-FA1922DFA9A8'
    """
    Contains magnetometer measurements for X, Y and Z axes as 3 signed 16 bit values in that order and in little endian 
    format. Data can be read on demand or notified periodically.
    
    Read Mandatory
    Notify Mandatory
    Fields
    1. Magnetometer_X : sint16
    2. Magnetometer_Y : sint16
    3. Magnetometer_Z : sint16
    Descriptors
    1. Client Characteristic Configuration : 2902
    """

    MAGNETOMETER_PERIOD = 'E95D386C-251D-470A-A062-FA1922DFA9A8'
    """
    Determines the frequency with which magnetometer data is reported in milliseconds.
    Valid values are 1, 2, 5, 10, 20, 80, 160 and 640.
    
    Read Mandatory
    Write Mandatory
    Fields
    1. Magnetometer_Period : uint16
    """

    MAGNETOMETER_BEARING = 'E95D9715-251D-470A-A062-FA1922DFA9A8'
    """
    Compass bearing in degrees from North.
    
    Read Mandatory
    Notify Mandatory
    Fields
    1. bearing value : uint16
    Descriptors
    1. Client Characteristic Configuration : 2902
    """

    MAGNETOMETER_CALIBRATION = 'E95DB358-251D-470A-A062-FA1922DFA9A8'
    """
    0 - state unknown
    1 - calibration requested
    2 - calibration completed OK
    3 - calibration completed with error
    
    Write Mandatory
    Notify Mandatory
    Fields
    1. calibration field : uint8
    Descriptors
    1. Client Characteristic Configuration : 2902
    """

    BUTTON_A = 'E95DDA90-251D-470A-A062-FA1922DFA9A8'
    """
    State of Button A may be read on demand by a connected client or the client may subscribe to notifications of 
    state change. 3 button states are defined and represented by a simple numeric enumeration: 
    0 = not pressed, 1 = pressed, 2 = long press.
    
    Read Mandatory
    Notify Mandatory
    Fields
    1. Button_State_Value : uint8
    Descriptors
    1. Client Characteristic Configuration : 2902
    """

    BUTTON_B = 'E95DDA91-251D-470A-A062-FA1922DFA9A8'
    """
    State of Button B may be read on demand by a connected client or the client may subscribe to notifications of 
    state change. 3 button states are defined and represented by a simple numeric enumeration:  
    0 = not pressed, 1 = pressed, 2 = long press.
    
    Read Mandatory
    Notify Mandatory
    Fields
    1. Button_State_Value : uint8
    Descriptors
    1. Client Characteristic Configuration : 2902
    """

    PIN_DATA = 'E95D8D00-251D-470A-A062-FA1922DFA9A8'
    """
    Contains data relating to zero or more pins. Structured as a variable length array of up to 
    19 Pin Number / Value pairs. 

    Pin Number and Value are each uint8 fields. 
    Note however that the micro:bit has a 10 bit ADC and so values are compressed to 8 bits with a loss of resolution.

    OPERATIONS:

    WRITE: Clients may write values to one or more pins in a single GATT write operation. 
    A pin to which a value is to be written must have been configured for output using the Pin IO Configuration 
    characteristic. Any attempt to write to a pin which is configured for input will be ignored.

    NOTIFY: Notifications will deliver Pin Number / Value pairs for those pins defined as input pins by the Pin IO 
    Configuration characteristic and whose value when read differs from the last read of the pin.

    READ: A client reading this characteristic will receive Pin Number / Value pairs for all those pins defined as 
    input pins by the Pin IO Configuration characteristic.
    
    Read Mandatory
    Write Mandatory
    Notify Mandatory
    Fields
    1. IO_Pin_Data : uint8[]
    Descriptors
    1. Client Characteristic Configuration : 2902
    """

    PIN_AD_CONFIGURATION = 'E95D5899-251D-470A-A062-FA1922DFA9A8'
    """
    A bit mask which allows each pin to be configured for analogue or digital use.

    Bit n corresponds to pin n where 0 LESS THAN OR EQUAL TO n LESS THAN 19. 
    A value of 0 means digital and 1 means analogue.
    
    Read Mandatory
    Write Mandatory
    Fields
    1. Pin_AD_Config_Value : uint24
    """

    PIN_IO_CONFIGURATION = 'E95DB9FE-251D-470A-A062-FA1922DFA9A8'
    """
    A bit mask (32 bit) which defines which inputs will be read. If the Pin AD Configuration bit mask is also set the 
    pin will be read as an analogue input, if not it will be read as a digital input.  

    Note that in practice, setting a pin's mask bit means that it will be read by the micro:bit runtime and, if 
    notifications have been enabled on the Pin Data characteristic, data read will be transmitted to the connected 
    Bluetooth peer device in a Pin Data notification. If the pin's bit is clear, it  simply means that it will not be 
    read by the micro:bit runtime.

    Bit n corresponds to pin n where 0 LESS THAN OR EQUAL TO n LESS THAN 19. A value of 0 means configured for output 
    and 1 means configured for input.
    
    Read Mandatory
    Write Mandatory
    Fields
    1. Pin_IO_Config_Value : uint24
    """

    PWM_CONTROL = 'E95DD822-251D-470A-A062-FA1922DFA9A8'
    """
    A variable length array 1 to 2 instances of :
    
    struct PwmControlData
     {
     uint8_t     pin;
     uint16_t    value;
     uint32_t    period;
    }
    
    Period is in microseconds and is an unsigned int but transmitted.
    Value is in the range 0 – 1024, per the current DAL API (e.g. setAnalogValue). 0 means OFF.
    
    Fields are transmitted over the air in Little Endian format.
    
        
    Write Mandatory
    Fields
    1. PWM Control Field : uint8[]
    """

    LED_MATRIX_STATE = 'E95D7B77-251D-470A-A062-FA1922DFA9A8'
    """
    Allows the state of any|all LEDs in the 5x5 grid to be set to on or off with a single GATT operation. 
    Consists of an array of 5 x utf8 octets, each representing one row of 5 LEDs.  
    Octet 0 represents the first row of LEDs i.e. the top row when the micro:bit is viewed with the edge connector at 
    the bottom and USB connector at the top. 
    Octet 1 represents the second row and so on.
    In each octet, bit 4 corresponds to the first LED in the row, bit 3 the second and so on. 
    Bit values represent the state of the related LED: off (0) or on (1).
    
    So we have:
    
    Octet 0, LED Row 1: bit4 bit3 bit2 bit1 bit0
    Octet 1, LED Row 2: bit4 bit3 bit2 bit1 bit0
    Octet 2, LED Row 3: bit4 bit3 bit2 bit1 bit0
    Octet 3, LED Row 4: bit4 bit3 bit2 bit1 bit0
    Octet 4, LED Row 5: bit4 bit3 bit2 bit1 bit0

    
    Read Mandatory
    Write Mandatory
    Fields
    1. LED_Matrix_State : uint8[]
    """

    LED_TEXT = 'E95D93EE-251D-470A-A062-FA1922DFA9A8'
    """
    A short UTF-8 string to be shown on the LED display. Maximum length 20 octets.
    
    Write Mandatory
    Fields
    1. LED_Text_Value : utf8s
    """

    SCROLLING_DELAY = 'E95D0D2D-251D-470A-A062-FA1922DFA9A8'
    """
    Specifies a millisecond delay to wait for in between showing each character on the display.
    
    Read Mandatory
    Write Mandatory
    Fields
    1. Scrolling_Delay_Value : uint16
    """

    MICROBIT_REQUIREMENTS = 'E95DB84C-251D-470A-A062-FA1922DFA9A8'
    """
    A variable length list of event data structures which indicates the types of client event, potentially with a 
    specific value which the micro:bit wishes to be informed of when they occur. The client should read this 
    characteristic when it first connects to the micro:bit. It may also subscribe to notifications
    to that it can be informed if the value of this characteristic is changed by the micro:bit firmware.
    
    struct event {
      uint16 event_type;
      uint16 event_value;
    };
    
    
    Note that an event_type of zero means ANY event type and an event_value part set to zero means ANY event value.
    
    event_type and event_value are each encoded in little endian format.
    
    Read Mandatory
    Notify Mandatory
    Fields
    1. microbit_reqs_value : uint8[]
    Descriptors
    1. Client Characteristic Configuration : 2902
    """

    MICROBIT_EVENT = 'E95D9775-251D-470A-A062-FA1922DFA9A8'
    """
    Contains one or more event structures which should be notified to the client. It supports notifications and as 
    such the client should subscribe to notifications from this characteristic.
    
    struct event {
      uint16 event_type;
      uint16 event_value;
    };
    
    
    Read Mandatory
    Notify Mandatory
    Fields
    1. Event_Type_And_Value : uint8[]
    Descriptors
    1. Client Characteristic Configuration : 2902
    """

    CLIENT_REQUIREMENTS = 'E95D23C4-251D-470A-A062-FA1922DFA9A8'
    """
    a variable length list of event data structures which indicates the types of micro:bit event, potentially with a 
    specific value which the client wishes to be informed of when they occur. The client should write to this 
    characteristic when it first connects to the micro:bit.
    
    struct event {
      uint16 event_type;
      uint16 event_value;
    };
    
    
    Note that an event_type of zero means ANY event type and an event_value part set to zero means ANY event value.
    
    event_type and event_value are each encoded in little endian format.

    Write Mandatory
    Fields
    1. Client_Requirements_Value : uint8[]
    Descriptors
    """

    CLIENT_EVENT = 'E95D5404-251D-470A-A062-FA1922DFA9A8'
    """
    a writable characteristic which the client may write one or more event structures to, to inform the micro:bit of 
    events which have occurred on the client. These should be of types indicated in the micro:bit Requirements 
    characteristic bit mask.
    
    struct event {
      uint16 event_type;
      uint16 event_value;
    };
    
    Write Mandatory
    Write Without Response Mandatory
    Fields
    1. Event_Types_And_Values : uint8[]
    Descriptors
    """

    DFU_CONTROL = 'E95D93B1-251D-470A-A062-FA1922DFA9A8'
    """
    Writing 0x01 initiates rebooting the micro:bit into the Nordic Semiconductor bootloader if the DFU Flash Code 
    characteristic has been written to with the correct secret key. 

    Writing 0x02 to this characteristic  means "request flash code".
    
    Read Mandatory
    Write Mandatory
    Fields
    1. dfu_control : uint8
    """

    TEMPERATURE = 'E95D9250-251D-470A-A062-FA1922DFA9A8'
    """
    Signed integer 8 bit value in degrees celsius.
    
    Read Mandatory
    Notify Mandatory
    Fields
    1. temperature value : sint8
    Descriptors
    1. Client Characteristic Configuration : 2902
    """

    TEMPERATURE_PERIOD = 'E95D1B25-251D-470A-A062-FA1922DFA9A8'
    """
    Determines the frequency with which temperature data is updated in milliseconds.
    
    Read Mandatory
    Write Mandatory
    Fields
    1. temperature period value : uint16
    Descriptors
    """

    TX_CHARACTERISTIC = '6E400002-B5A3-F393-E0A9-E50E24DCCA9E'
    """
    This characteristic allows the micro:bit to transmit a byte array containing an arbitrary number of arbitrary 
    octet values to a connected device. 

    The maximum number of bytes which may be transmitted in one PDU is limited to the MTU minus three or 20 octets 
    to be precise.

    Indicate Mandatory
    Fields
    1. UART TX Field : uint8[]
    Descriptors
    """

    RX_CHARACTERISTIC = '6E400003-B5A3-F393-E0A9-E50E24DCCA9E'
    """
    This characteristic allows a connected client to send a byte array containing an arbitrary number of arbitrary 
    octet values to a connected micro:bit. 

    The maximum number of bytes which may be transmitted in one PDU is limited to the MTU minus three or 20 octets 
    to be precise.
    
    
    Write Mandatory
    Write Without Response Mandatory
    Fields
    1. UART TX Field : uint8[]

    """



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Union
from bleak import BleakClient
import asyncio
from threading import Thread
from .characteristics import Characteristic

ByteData = Union[bytes, bytearray, memoryview]


class ThreadEventLoop:
    def __init__(self):
        self.loop = asyncio.new_event_loop()
        Thread(target=ThreadEventLoop._start_background_loop, args=(self.loop,), daemon=True).start()

    @staticmethod
    def _start_background_loop(loop: asyncio.AbstractEventLoop) -> None:
        asyncio.set_event_loop(loop)
        loop.run_forever()

    def run_async(self, coroutine):
        return asyncio.run_coroutine_threadsafe(coroutine, self.loop)


class BluetoothEventLoop:
    _single_thread = ThreadEventLoop()

    @staticmethod
    def single_thread():
        return BluetoothEventLoop._single_thread


class BluetoothDevice:

    def __init__(self, address: str, loop=BluetoothEventLoop.single_thread()):
        self.loop = loop
        self.client = BleakClient(address)

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self) -> None:
        print("connecting")
        self.loop.run_async(self.client.connect()).result()
        print("connected")

    def disconnect(self) -> None:
        print("disconnecting")
        self.loop.run_async(self.client.disconnect()).result()
        print("disconnected")

    def read(self, characteristic: Characteristic) -> bytearray:
        print("reading")
        result = self.loop.run_async(self.client.read_gatt_char(characteristic.value)).result()
        print("read")
        return result

    def write(self, characteristic: Characteristic, data: ByteData) -> None:
        print("writing")
        self.loop.run_async(self.client.write_gatt_char(characteristic.value, data)).result()
        print("written")

    def notify(self, characteristic: Characteristic, callback) -> None:
        self.loop.run_async(self.client.start_notify(characteristic.value, callback)).result()
        print("notify")



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at https://mozilla.org/MPL/2.0/.

from .bluetoothdevice import BluetoothDevice, BluetoothEventLoop
from .services.device_information import DeviceInformationService
from .services.generic_access import GenericAccessService
from .services.buttons import ButtonService
from .services.temperature import TemperatureService
from .services.accelerometer import AccelerometerService
from .services.events import EventService
from .services.uart import UartService
from .services.magnetometer import MagnetometerService
from .services.io_pin import IOPinService
from .services.led import LedService


class KaspersMicrobit:
    def __init__(self, address: str, loop=BluetoothEventLoop.single_thread()):
        self._device = BluetoothDevice(address, loop=loop)
        self.device_information = DeviceInformationService(self._device)
        self.generic_access = GenericAccessService(self._device)
        self.buttons = ButtonService(self._device)
        self.temperature = TemperatureService(self._device)
        self.accelerometer = AccelerometerService(self._device)
        self.events = EventService(self._device)
        self.uart = UartService(self._device)
        self.io_pin = IOPinService(self._device)
        self.led = LedService(self._device)
        self.magnetometer = MagnetometerService(self._device)

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self) -> None:
        self._device.connect()

    def disconnect(self) -> None:
        self._device.disconnect()

